// tooling
const browserslist = require('browserslist');
const cssdb = require('cssdb');
const postcss = require('postcss');
const plugins = require('./lib/plugins-by-specification-id');
const getTransformedInsertions = require('./lib/get-transformed-insertions');
const getUnsupportedBrowsersByFeature = require('./lib/get-unsupported-browsers-by-feature');
const specificationIdsByExecutionOrder = require('./lib/specification-ids-by-execution-order');

// postcss-preset-env
module.exports = postcss.plugin('postcss-preset-env', rawopts => {
	// options
	const opts = Object(rawopts);
	const stage = 'stage' in opts ? opts.stage === false ? 6 : parseInt(opts.stage) || 0 : 3;
	const features = Object(opts.features);
	const insertBefore = Object(opts.insertBefore);
	const insertAfter = Object(opts.insertAfter);

	// polyfillable features (those with an available postcss plugin)
	const polyfillableFeatures = cssdb.concat(
		// features to be inserted before cssdb features
		getTransformedInsertions(insertBefore, 'insertBefore'),
		// features to be inserted after cssdb features
		getTransformedInsertions(insertAfter, 'insertAfter')
	).sort(
		(a, b) => specificationIdsByExecutionOrder.indexOf(a.specificationId) - specificationIdsByExecutionOrder.indexOf(b.specificationId) || (a.insertBefore ? -1 : b.insertBefore ? 1 : 0) || (a.insertAfter ? 1 : b.insertAfter ? -1 : 0)
	).filter(feature => feature.insertBefore || feature.specificationId in plugins).map(feature => {
		const unsupportedBrowsers = getUnsupportedBrowsersByFeature(feature.caniuse);

		return feature.insertBefore || feature.insertAfter ? {
			browsers:        unsupportedBrowsers,
			plugin:          feature.plugin,
			specificationId: `${feature.insertBefore ? 'before' : 'after'}-${feature.specificationId}`,
			stage:           6
		} : {
			browsers:        unsupportedBrowsers,
			plugin:          plugins[feature.specificationId],
			specificationId: feature.specificationId,
			stage:           feature.stage
		};
	});

	// staged features (those at or above the selected stage)
	const stageSupportedFeatures = polyfillableFeatures.filter(feature => feature.specificationId in features ? features[feature.specificationId] : feature.stage >= stage).map(
		feature => ({
			browsers: feature.browsers,
			plugin: typeof feature.plugin.process === 'function' ? feature.plugin(features[feature]) : feature.plugin,
			specificationId: feature.specificationId
		})
	);

	return (root, result) => {
		// client browser list
		const clientBrowserList = browserslist(opts.browsers, {
			path: root.source && root.source.input && root.source.input.file
		});

		// list of stage-supported, browser-supported polyfills
		const browserSupportedFeatures = stageSupportedFeatures.filter(
			feature => clientBrowserList.some(
				clientBrowser => browserslist(feature.browsers).some(polyfillBrowser => polyfillBrowser === clientBrowser)
			)
		);

		// run all polyfills
		return browserSupportedFeatures.reduce(
			(promise, feature) => promise.then(
				() => feature.plugin(root, result)
			),
			Promise.resolve()
		);
	};
});

